"""Data model for homeegrams (automations) in Homee."""

from datetime import datetime, timezone
from urllib.parse import unquote
from typing import Any

from pyHomee.model import HomeeObject


class HomeegramTcaItem:
    # TCA stands for trigger, condition, action. This class is used as a base class for these.
    """Representation of a homeegram trigger."""

    def __init__(self, tca_dict: dict[str, Any]) -> None:
        """Initialize a tcaItem."""
        self._data = tca_dict

    @property
    def data(self) -> dict[str, Any]:
        """Return the raw tcaItem dict."""
        return dict(self._data)

    @property
    def id(self) -> int:
        """Id of the tcaItem, unique in Homee."""
        return int(self._data["id"])

    @property
    def homeegram_id(self) -> int:
        """Id of the homeegram the tcaItem belongs to."""
        return int(self._data["homeegram_id"])


class HomeegramTriggers:
    """Representation of homeegram triggers."""

    class HomeegramTimeTrigger(HomeegramTcaItem):
        """Representation of a homeegram time trigger."""

        @property
        def dtstart(self) -> datetime:
            """Start date/time of the time trigger."""
            return datetime.strptime(self._data["dtstart"], "%Y%m%dT%H%M%SZ").replace(
                tzinfo=timezone.utc
            )

        @property
        def rrule(self) -> str:
            """Recurrence rule of the time trigger."""
            return str(self._data["rrule"])

        @property
        def next_invoaction(self) -> datetime | None:
            """Time of the next invocation of the time trigger."""
            if "next_invocation" in self._data:
                return datetime.strptime(
                    self._data["next_invocation"], "%Y-%m-%dT%H:%M:%S"
                )
            else:
                return None

    class HomeegramAttributeTrigger(HomeegramTcaItem):
        """Representation of a homeegram attribute trigger."""

        @property
        def node_id(self) -> int:
            """Id of the node the triggering attribute belongs to."""
            return int(self._data["node_id"])

        @property
        def attribute_id(self) -> int:
            """Id of the triggering attribute."""
            return int(self._data["attribute_id"])

        @property
        def operator(self) -> int:
            """Operator used in the trigger."""
            return int(self._data["operator"])

        @property
        def operand(self) -> int:
            """Operand used in the trigger."""
            return int(self._data["operand"])

        @property
        def value(self) -> float:
            """Value of the trigger."""
            return float(self._data["value"])

    class HomeegramWebhookTrigger(HomeegramTcaItem):
        """Representation of a homeegram webhook trigger."""

        @property
        def event(self) -> str:
            """Event name of the webhook trigger."""
            return str(self._data["event"])

    class HomeegramCelestialTrigger(HomeegramTcaItem):
        """Representation of a homeegram celestial trigger."""

        @property
        def celestial_type(self) -> int:
            """Type of celestial event triggering this trigger."""
            return int(self._data["celestial_type"])

        @property
        def time_offset(self) -> int:
            """Time offset in minutes for the celestial trigger."""
            return int(self._data["time_offset"])

    class HomeegramGroupTrigger(HomeegramTcaItem):
        """Representation of a homeegram group trigger."""

        @property
        def group_id(self) -> int:
            """Id of the group triggering this trigger."""
            return int(self._data["group_id"])

        @property
        def attribute_type(self) -> int:
            """Attribute type of the group trigger."""
            return int(self._data["attribute_type"])

        @property
        def operator(self) -> int:
            """Operator used in the trigger."""
            return int(self._data["operator"])

        @property
        def value(self) -> float:
            """Value of the trigger."""
            return float(self._data["value"])

    class HomeegramUserTrigger(HomeegramTcaItem):
        """Representation of a homeegram user trigger."""

        @property
        def user_id(self) -> int:
            """Id of the user triggering this trigger."""
            return int(self._data["user_id"])

        @property
        def value(self) -> int:
            """Value of the user trigger."""
            return int(self._data["value"])

        @property
        def data(self) -> dict[str, list]:
            """Return the raw triggers dict."""
            return dict(self._data)

    def __init__(self, triggers: dict[str, list[dict[str, Any]]]) -> None:
        """Initialize triggers."""
        self._data = triggers

    @property
    def data(self) -> dict[str, list[dict[str, Any]]]:
        """Return the raw triggers dict."""
        return dict(self._data)

    @property
    # A switch trigger has no additional properties.
    def switch_triggers(self) -> list[HomeegramTcaItem]:
        """Return the list of switch triggers."""
        return [HomeegramTcaItem(trigger) for trigger in self._data["switch_triggers"]]

    @property
    def time_triggers(self) -> list[HomeegramTimeTrigger]:
        """Return the list of time triggers."""
        return [
            self.HomeegramTimeTrigger(trigger)
            for trigger in self._data["time_triggers"]
        ]

    @property
    def attribute_triggers(self) -> list[HomeegramAttributeTrigger]:
        """Return the list of attribute triggers."""
        return [
            self.HomeegramAttributeTrigger(trigger)
            for trigger in self._data["attribute_triggers"]
        ]

    @property
    def webhook_triggers(self) -> list[HomeegramWebhookTrigger]:
        """Return the list of webhook triggers."""
        return [
            self.HomeegramWebhookTrigger(trigger)
            for trigger in self._data["webhook_triggers"]
        ]

    @property
    # For homeegram triggers I have no information about the structure of
    # the trigger dict, so I return a list of basic HomeegramTcaItems.
    def homeegram_triggers(self) -> list[HomeegramTcaItem]:
        """Return the list of homeegram triggers."""
        return [
            HomeegramTcaItem(trigger) for trigger in self._data["homeegram_triggers"]
        ]

    @property
    def celestial_triggers(self) -> list[HomeegramCelestialTrigger]:
        """Return the list of celestial triggers."""
        return [
            self.HomeegramCelestialTrigger(trigger)
            for trigger in self._data["celestial_triggers"]
        ]

    @property
    # For plan triggers I have no information about the structure of
    # the trigger dict, so I return a list of basic HomeegramTcaItems.
    def plan_triggers(self) -> list[HomeegramTcaItem]:
        """Return the list of plan triggers."""
        return [HomeegramTcaItem(trigger) for trigger in self._data["plan_triggers"]]

    @property
    def group_triggers(self) -> list[HomeegramGroupTrigger]:
        """Return the list of group triggers."""
        return [
            self.HomeegramGroupTrigger(trigger)
            for trigger in self._data["group_triggers"]
        ]

    @property
    def user_triggers(self) -> list[HomeegramUserTrigger]:
        """Return the list of user triggers."""
        return [
            self.HomeegramUserTrigger(trigger)
            for trigger in self._data["user_triggers"]
        ]


class HomeegramConditions:
    """Representation of homeegram conditions."""

    class HomeegramCondition(HomeegramTcaItem):
        """Representation of a homeegram condition."""

        @property
        def check_moment(self) -> int:
            """Check moment of the condition."""
            return int(self._data["check_moment"])

    class HomeegramTimeCondition(HomeegramCondition):
        """Representation of a homeegram time condition."""

        @property
        def dtstart(self) -> datetime:
            """Start date/time of the time condition."""
            return datetime.strptime(self._data["dtstart"], "%Y%m%dT%H%M%SZ").replace(
                tzinfo=timezone.utc
            )

        @property
        def rrule(self) -> str:
            """Recurrence rule of the time condition."""
            return str(self._data["rrule"])

        @property
        def duration(self) -> str:
            """Duration of the time condition."""
            return str(self._data["duration"])

    class HomeegramAttributeCondition(HomeegramCondition):
        """Representation of a homeegram attribute condition."""

        @property
        def node_id(self) -> int:
            """Id of the node the condition attribute belongs to."""
            return int(self._data["node_id"])

        @property
        def attribute_id(self) -> int:
            """Id of the condition attribute."""
            return int(self._data["attribute_id"])

        @property
        def operator(self) -> int:
            """Operator used in the condition."""
            return int(self._data["operator"])

        @property
        def operand(self) -> int:
            """Operand used in the condition."""
            return int(self._data["operand"])

        @property
        def value(self) -> float:
            """Value of the condition."""
            return float(self._data["value"])

    class HomeegramHomeegramCondition(HomeegramCondition):
        """Representation of a homeegram homeegram condition."""

        @property
        def source_homeegram_id(self) -> int:
            """Id of the source homeegram for this condition."""
            return int(self._data["source_homeegram_id"])

        @property
        def homeegram_event(self) -> int:
            """Homeegram event type for this condition."""
            return int(self._data["homeegram_event"])

    class HomeegramCelestialCondition(HomeegramCondition):
        """Representation of a homeegram celestial condition."""

        @property
        def time_of_day(self) -> int:
            """Time of day for this celestial condition."""
            return int(self._data["time_of_day"])

    class HomeegramGroupCondition(HomeegramCondition):
        """Representation of a homeegram group condition."""

        @property
        def group_id(self) -> int:
            """Id of the group for this condition."""
            return int(self._data["group_id"])

        @property
        def attribute_type(self) -> int:
            """Attribute type of the group condition."""
            return int(self._data["attribute_type"])

        @property
        def operator(self) -> int:
            """Operator used in the condition."""
            return int(self._data["operator"])

        @property
        def value(self) -> float:
            """Value of the condition."""
            return float(self._data["value"])

        @property
        def scope(self) -> int:
            """Scope of the group condition."""
            return int(self._data["scope"])

    class HomeegramUserCondition(HomeegramCondition):
        """Representation of a homeegram user condition."""

        @property
        def user_id(self) -> int:
            """Id of the user for this condition."""
            return int(self._data["user_id"])

        @property
        def value(self) -> int:
            """Value of the user condition."""
            return int(self._data["value"])

    def __init__(self, conditions: dict[str, Any]) -> None:
        """Initialize conditions."""
        self._data = conditions

    @property
    def data(self) -> dict[str, list]:
        """Return the raw conditions dict."""
        return dict(self._data)

    @property
    def time_conditions(self) -> list[HomeegramTimeCondition]:
        """Return the list of time conditions."""
        return [
            self.HomeegramTimeCondition(condition)
            for condition in self._data["time_conditions"]
        ]

    @property
    def attribute_conditions(self) -> list[HomeegramAttributeCondition]:
        """Return the list of attribute conditions."""
        return [
            self.HomeegramAttributeCondition(condition)
            for condition in self._data["attribute_conditions"]
        ]

    @property
    def homeegram_conditions(self) -> list[HomeegramTcaItem]:
        # For homeegram conditions I have no information about the structure of
        # the condition dict, so I return a list of basic HomeegramTcaItems.
        """Return the list of homeegram conditions."""
        return [
            HomeegramTcaItem(condition)
            for condition in self._data["homeegram_conditions"]
        ]

    @property
    def celestial_conditions(self) -> list[HomeegramCelestialCondition]:
        """Return the list of celestial conditions."""
        return [
            self.HomeegramCelestialCondition(condition)
            for condition in self._data["celestial_conditions"]
        ]

    @property
    def plan_conditions(self) -> list[HomeegramTcaItem]:
        # For plan conditions I have no information about the structure of
        # the condition dict, so I return a list of basic HomeegramTcaItems.
        """Return the list of plan conditions."""
        return [
            HomeegramTcaItem(condition) for condition in self._data["plan_conditions"]
        ]

    @property
    def group_conditions(self) -> list[HomeegramGroupCondition]:
        """Return the list of group conditions."""
        return [
            self.HomeegramGroupCondition(condition)
            for condition in self._data["group_conditions"]
        ]

    @property
    def user_conditions(self) -> list[HomeegramUserCondition]:
        """Return the list of user conditions."""
        return [
            self.HomeegramUserCondition(condition)
            for condition in self._data["user_conditions"]
        ]


class HomeegramActions:
    """Representation of homeegram actions."""

    class HomeegramAction(HomeegramTcaItem):
        """Representation of a homeegram action."""

        @property
        def delay(self) -> int:
            """Delay in seconds before the action is executed."""
            return int(self._data["delay"])

    class HomeegramAttributeAction(HomeegramAction):
        """Representation of a homeegram attribute action."""

        @property
        def node_id(self) -> int:
            """Id of the node the action attribute belongs to."""
            return int(self._data["node_id"])

        @property
        def attribute_id(self) -> int:
            """Id of the action attribute."""
            return int(self._data["attribute_id"])

        @property
        def source_attribute_id(self) -> int:
            """Id of the source attribute."""
            return int(self.data["source_attribute_id"])

        @property
        def value(self) -> float:
            """Value of the action."""
            return float(self._data["value"])

        @property
        def command(self) -> int:
            """Command type associated with the action."""
            return int(self._data["command"])

    class HomeegramGroupAction(HomeegramAction):
        """Representation of a homeegram group action."""

        @property
        def group_id(self) -> int:
            """Id of the group the action belongs to."""
            return int(self._data["group_id"])

        @property
        def attribute_type(self) -> int:
            """Attribute type of the group action."""
            return int(self._data["attribute_type"])

        @property
        def source_attribute_id(self) -> int:
            """Id of the source attribute."""
            return int(self._data["source_attribute_id"])

        @property
        def value(self) -> float:
            """Value of the action."""
            return float(self._data["value"])

        @property
        def command(self) -> int:
            """Command type associated with the action."""
            return int(self._data["command"])

    class HomeegramNotificationAction(HomeegramAction):
        """Representation of a homeegram notification action."""

        @property
        def style(self) -> int:
            """Style of the notification."""
            return int(self._data["style"])

        @property
        def critical(self) -> bool:
            """Whether the notification is critical."""
            return bool(self._data["critical"])

        @property
        def user_ids(self) -> list[int]:
            """List of user ids to notify."""
            return [int(user_id) for user_id in self._data["user_ids"]]

        @property
        def message(self) -> str:
            """Decoded message of the notification."""
            return unquote(self._data["message"])

    class HomeegramWebhookAction(HomeegramAction):
        """Representation of a homeegram webhook action."""

        @property
        def method(self) -> str:
            """HTTP method of the webhook action."""
            return str(self._data["method"])

        @property
        def url(self) -> str:
            """Decoded URL of the webhook action."""
            return unquote(self._data["url"])

        @property
        def body(self) -> str:
            """Body of the webhook action."""
            return str(self._data["body"])

        @property
        def content_type(self) -> str:
            """Content type of the webhook action."""
            return str(self._data["content_type"])

    class HomeegramHomeegramAction(HomeegramAction):
        """Representation of a homeegram homeegram action."""

        @property
        def target_homeegram_id(self) -> int:
            """Id of the target homeegram for this action."""
            return int(self._data["target_homeegram_id"])

        @property
        def homeegram_event(self) -> int:
            """Homeegram event type for this action."""
            return int(self._data["homeegram_event"])

    def __init__(self, actions: dict[str, Any]) -> None:
        """Initialize actions."""
        self._data = actions

    @property
    def data(self) -> dict[str, list]:
        """Return the raw actions dict."""
        return dict(self._data)

    @property
    def attribute_actions(self) -> list[HomeegramAttributeAction]:
        """Return the list of attribute actions."""
        return [
            self.HomeegramAttributeAction(action)
            for action in self._data["attribute_actions"]
        ]

    @property
    def group_action(self) -> list[HomeegramGroupAction]:
        """Return the list of group actions."""
        return [
            self.HomeegramGroupAction(action) for action in self._data["group_actions"]
        ]

    @property
    def tts_actions(self) -> list[HomeegramTcaItem]:
        # For tts actions I have no information about the structure of
        # the action dict, so I return a list of basic HomeegramTcaItems.
        """Retrun the list of tts actions"""
        return [HomeegramTcaItem(action) for action in self._data["tts_actions"]]

    @property
    def notification_actions(self) -> list[HomeegramNotificationAction]:
        """Return the list of notification actions."""
        return [
            self.HomeegramNotificationAction(action)
            for action in self._data["notification_actions"]
        ]

    @property
    def webhook_actions(self) -> list[HomeegramWebhookAction]:
        """Return the list of webhook actions."""
        return [
            self.HomeegramWebhookAction(action)
            for action in self._data["webhook_actions"]
        ]

    @property
    def homeegram_actions(self) -> list[HomeegramHomeegramAction]:
        """Return the list of homeegram actions."""
        return [
            self.HomeegramHomeegramAction(action)
            for action in self._data["homeegram_actions"]
        ]

    @property
    def plan_actions(self) -> list[HomeegramTcaItem]:
        # For plan actions I have no information about the structure of
        # the action dict, so I return a list of basic HomeegramTcaItems.
        """Retrun the list of plan actions"""
        return [HomeegramTcaItem(action) for action in self._data["plan_actions"]]

    @property
    def user_actions(self) -> list[HomeegramTcaItem]:
        # For user actions I have no information about the structure of
        # the action dict, so I return a list of basic HomeegramTcaItems.
        """Retrun the list of user actions"""
        return [HomeegramTcaItem(action) for action in self._data["user_actions"]]


class HomeeGram(HomeeObject):
    """Representation of a homeegram (automation)."""

    @property
    def id(self) -> int:
        """Id of the homeegram, unique in homee."""
        return int(self._data["id"])

    @property
    def name(self) -> str:
        """Decoded user given name of the homeegram."""
        return unquote(self._data["name"])

    @property
    def image(self) -> str:
        """Image identifier of the homeegram."""
        return unquote(self._data["image"])

    @property
    def state(self) -> int:
        """State of the homeegram."""
        return int(self._data["state"])

    @property
    def visible(self) -> bool:
        """Whether the homeegram is visible."""
        return bool(self._data["visible"])

    @property
    def favorite(self) -> bool:
        """Whether the homeegram is marked as favorite."""
        return bool(self._data["favorite"])

    @property
    def order(self) -> int:
        """Order of the homeegram in the list."""
        return int(self._data["order"])

    @property
    def active(self) -> bool:
        """Whether the homeegram is active."""
        return bool(self._data["active"])

    @property
    def play(self) -> bool:
        """Whether the homeegram is currently playing."""
        return bool(self._data["play"])

    @property
    def added(self) -> int:
        """Timestamp the homeegram was added."""
        return int(self._data["added"])

    @property
    def phonetic_name(self) -> str:
        """Phonetic name of the homeegram."""
        return unquote(self._data["phonetic_name"])

    @property
    def note(self) -> str:
        """Note describing the homeegram."""
        return unquote(self._data["note"])

    @property
    def services(self) -> int:
        """Services associated with the homeegram."""
        return int(self._data["services"])

    @property
    def last_triggered(self) -> int:
        """Time the homeegram was last triggered."""
        return int(self._data["last_triggered"])

    @property
    def owner(self) -> int:
        """Owner id of the homeegram."""
        return int(self._data["owner"])

    @property
    def triggers(self) -> HomeegramTriggers:
        """Dict containing all triggers of the homeegram."""
        return HomeegramTriggers(self._data["triggers"])

    @property
    def conditions(self) -> HomeegramConditions:
        """Dict containing all conditions of the homeegram."""
        return HomeegramConditions(self._data["conditions"])

    @property
    def actions(self) -> HomeegramActions:
        """Dict containing all actions of the homeegram."""
        return HomeegramActions(self._data["actions"])



"""Library for interacting with the homee smart home/home automation platform."""

import asyncio
from collections.abc import Callable, Coroutine
from datetime import datetime
import hashlib
import json
import logging
from urllib.parse import parse_qs
from typing import Any, Literal

import aiohttp
import aiohttp.client_exceptions
from aiohttp.helpers import BasicAuth
import websockets.asyncio.client
import websockets.exceptions

from .const import DeviceApp, DeviceOS, DeviceType, WarningCode
from .model import (
    HomeeDevice,
    HomeeGroup,
    HomeeNode,
    HomeeRelationship,
    HomeeSettings,
    HomeeUser,
    HomeeWarning,
)
from .model_homeegram import HomeeGram

_LOGGER = logging.getLogger(__name__)


class Homee:
    """Representation of a Homee system."""

    def __init__(
        self,
        host: str,
        user: str,
        password: str,
        device: str = "pymee",
        ping_interval: int = 30,
        reconnect_interval: int = 5,
        reconnect: bool = True,
        max_retries: int = 5,
    ) -> None:
        """Initialize the virtual Homee."""
        self.host: str = host
        self.user: str = user
        self.password: str = password

        self.device: str = device
        self.ping_interval: int = ping_interval
        self.should_reconnect: bool = reconnect
        self.reconnect_interval: int = reconnect_interval
        self.max_retries: int = max_retries

        self.device_id: str = str(device).lower().replace(" ", "-")

        self.devices: list[HomeeDevice] = []
        self.groups: list[HomeeGroup] = []
        self.nodes: list[HomeeNode] = []
        self.relationships: list[HomeeRelationship] = []
        self.settings: HomeeSettings
        self.users: list[HomeeUser] = []
        self.homeegrams: list[HomeeGram] = []
        self.warning = HomeeWarning(
            {"code": 0, "description": "None", "message": "None", "data": {}}
        )
        self.token: str = ""
        self.expires: float = 0
        self.connected: bool = False
        self.retries: int = 0
        self.should_close: bool = False

        self._message_queue: asyncio.Queue[str] = asyncio.Queue()
        self._connected_event = asyncio.Event()
        self._disconnected_event = asyncio.Event()
        self._connection_listeners: list[
            Callable[[bool], Coroutine[Any, Any, None]]
        ] = []
        self._nodes_listeners: list[
            Callable[[HomeeNode, bool], Coroutine[Any, Any, None]]
        ] = []

    async def get_access_token(self) -> str:
        """Try asynchronously to get an access token from homee using username and password."""

        # Check if current token is still valid
        if self.token is not None and self.expires > datetime.now().timestamp():
            return self.token

        auth = BasicAuth(
            self.user, hashlib.sha512(self.password.encode("utf-8")).hexdigest()
        )
        url = f"{self.url}/access_token"
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        data = {
            "device_name": self.device,
            "device_hardware_id": self.device_id,
            "device_os": DeviceOS.LINUX,
            "device_type": DeviceType.NONE,
            "device_app": DeviceApp.HOMEE,
        }

        try:
            async with aiohttp.ClientSession() as client:
                req = await client.post(
                    url,
                    auth=auth,
                    data=data,
                    headers=headers,
                    timeout=aiohttp.ClientTimeout(total=5),
                )

                try:
                    req_text = await req.text()
                except (
                    aiohttp.client_exceptions.ClientError,
                    asyncio.TimeoutError,
                    UnicodeDecodeError,
                    LookupError,
                ) as e:
                    raise HomeeAuthFailedException(
                        f"Failed to decode response: {e}"
                    ) from e
        except asyncio.TimeoutError as e:
            raise HomeeConnectionFailedException("Connection to Homee timed out") from e
        except aiohttp.client_exceptions.ClientError as e:
            raise HomeeConnectionFailedException("Could not connect to Homee") from e

        if req.status == 200:
            try:
                parsed = parse_qs(req_text, strict_parsing=True)
                self.token = parsed["access_token"][0]
                expires = int(parsed["expires"][0])
                self.expires = datetime.now().timestamp() + expires
                self.retries = 0
            except (KeyError, IndexError, ValueError) as e:
                raise HomeeAuthFailedException(
                    f"Invalid token format: {req_text}"
                ) from e
        else:
            raise HomeeAuthFailedException(
                f"Auth request was unsuccessful. Status: {req.status} - {req.reason}"
            )

        return self.token

    async def run(self) -> None:
        """Connect to homee after acquiring an access token and runs until the connection is closed.

        Should be used in combination with asyncio.create_task.
        """

        self.should_close = False
        initial_connect = True

        # Reconnect loop to avoid recursive reconnects
        while initial_connect or (
            not self.should_close
            and self.should_reconnect
            and self.retries < self.max_retries
        ):
            initial_connect = False

            # Sleep after reconnect
            if self.retries > 0:
                await asyncio.sleep(self.reconnect_interval * self.retries)
                _LOGGER.debug(
                    "Attempting to reconnect in %s seconds",
                    self.reconnect_interval * self.retries,
                )

            try:
                await self.get_access_token()
            except HomeeConnectionFailedException as e:
                _LOGGER.debug("Could not connect to Homee: %s", e)
                # Reconnect
                self.retries += 1
                continue
            except HomeeAuthFailedException as e:
                _LOGGER.debug("Could not authenticate with Homee: %s", e)
                # Do not reconnect, since the authentication will not magically work.
                self.should_reconnect = False
                continue

            await self.open_ws()

        # Handle max retries
        if self.retries >= self.max_retries:
            await self.on_max_retries()

    def start(self) -> asyncio.Task:
        """Wrap run() with asyncio.create_task() and returns the resulting task."""
        return asyncio.create_task(self.run())

    async def open_ws(self) -> None:
        """Open the websocket connection assuming an access token was already received.

        Runs until connection is closed again.
        """

        _LOGGER.debug("Opening websocket")

        if self.retries > 0:
            await self.on_reconnect()

        try:
            async with websockets.asyncio.client.connect(
                uri=f"{self.ws_url}/connection?access_token={self.token}",
                subprotocols=[websockets.Subprotocol("v2")],
                # homee sends a large "all" dump; avoid default limit
                max_size=None,
            ) as ws:
                await self._ws_on_open()

                while (not self.should_close) and self.connected:
                    try:
                        receive_task = asyncio.ensure_future(
                            self._ws_receive_handler(ws)
                        )
                        send_task = asyncio.ensure_future(self._ws_send_handler(ws))
                        done, pending = await asyncio.wait(
                            [receive_task, send_task],
                            return_when=asyncio.FIRST_COMPLETED,
                        )

                        exceptions: list[BaseException | None] = []

                        # Kill pending tasks
                        for task in pending:
                            task.cancel()

                        # Check if we finished with an exception
                        exceptions.extend(task.exception() for task in done)

                        if exceptions and exceptions[0] is not None:
                            raise exceptions[0]

                    except websockets.exceptions.ConnectionClosedError as e:
                        self.connected = False
                        await self.on_disconnected(e)
        except (
            websockets.exceptions.WebSocketException,
            ConnectionError,
            OSError,
        ) as e:
            await self._ws_on_error(e)

        self.retries += 1
        await self._ws_on_close()

    async def _ws_receive_handler(
        self, ws: websockets.asyncio.client.ClientConnection
    ) -> None:
        try:
            msg = await ws.recv(decode=True)
            await self._ws_on_message(msg)
        except websockets.exceptions.ConnectionClosedOK:
            return
        except websockets.exceptions.ConnectionClosedError as e:
            if not self.should_close:
                self.connected = False
                raise e

    async def _ws_send_handler(
        self, ws: websockets.asyncio.client.ClientConnection
    ) -> None:
        """Continuously send messages from the message queue until the connection is closed."""
        try:
            msg = await self._message_queue.get()
            if self.connected and not self.should_close:
                await ws.send(msg)
        except websockets.exceptions.ConnectionClosed as e:
            if not self.should_close:
                self.connected = False
                raise e

    async def _ws_on_open(self) -> None:
        """Websocket on_open callback."""

        _LOGGER.debug("Connection to websocket successful")

        self.connected = True

        await self.on_connected()
        await self.send("GET:all")

    async def _ws_on_message(self, msg: str) -> None:
        """Websocket on_message callback."""

        await self._handle_message(json.loads(msg))

    async def _ws_on_close(self) -> None:
        """Websocket on_close callback."""
        # if not self.should_close and self.retries <= 1:

        if self.connected:
            self.connected = False
            self._disconnected_event.set()

            await self.on_disconnected()

    async def _ws_on_error(self, error: Exception) -> None:
        """Websocket on_error callback."""

        await self.on_error(error)

    async def send(self, msg: str) -> None:
        """Send a raw string message to homee."""

        if not self.connected or self.should_close:
            return

        await self._message_queue.put(msg)

    async def reconnect(self) -> None:
        """Start a reconnection attempt."""

        await self.run()

    def disconnect(self) -> None:
        """Disconnect from homee by closing the websocket connection."""

        self.should_close = True

    def add_connection_listener(
        self, listener: Callable[[bool], Coroutine[Any, Any, None]]
    ) -> Callable[[], None]:
        """Add a listener for change in connected state."""
        self._connection_listeners.append(listener)

        def remove_listener() -> None:
            self._connection_listeners.remove(listener)

        return remove_listener

    def add_nodes_listener(
        self, listener: Callable[[HomeeNode, bool], Coroutine[Any, Any, None]]
    ) -> Callable[[], None]:
        """Add a listener for node add/delete events."""
        self._nodes_listeners.append(listener)

        def remove_listener() -> None:
            self._nodes_listeners.remove(listener)

        return remove_listener

    async def _handle_message(self, msg: dict) -> None:
        """Handle incoming homee messages."""

        msg_type = None

        try:
            msg_type = list(msg)[0]
        except TypeError as e:
            _LOGGER.info("Invalid message: %s", msg)
            await self.on_error(e)
            return

        _LOGGER.debug(msg)

        if msg_type == "all":
            self.settings = HomeeSettings(msg["all"]["settings"])

            # Create / Update nodes
            if not self.nodes:
                # Since there might be lots of nodes, we don't want to check for
                # all in the next step, so if we start up, just add all nodes.
                self.nodes = [HomeeNode(node_data) for node_data in msg["all"]["nodes"]]
            else:
                for node_data in msg["all"]["nodes"]:
                    await self._update_or_create_node(node_data, self.warning.code)

            # Create / Update groups
            for group_data in msg["all"]["groups"]:
                self._update_or_create_group(group_data)

            # Create / Update users
            for user_data in msg["all"]["users"]:
                self._update_or_create_user(user_data)

            # Create / Update homeegrams
            for homeegram_data in msg["all"]["homeegrams"]:
                self._update_or_create_homeegram(homeegram_data)

            # Create / Update relationships
            self._update_or_create_relationships(msg["all"]["relationships"])
            self._remap_relationships()

            self._connected_event.set()

        elif msg_type == "attribute":
            await self._handle_attribute_change(msg["attribute"])
        # Not sure, if devices can be sent alone or only with user, but just in case...
        elif msg_type == "device":
            self._update_or_create_device(msg["device"])
        elif msg_type == "devices":
            for data in msg["devices"]:
                self._update_or_create_device(data)
        elif msg_type == "group":
            self._update_or_create_group(msg["group"])
        elif msg_type == "groups":
            for data in msg["groups"]:
                self._update_or_create_group(data)
        elif msg_type == "node":
            if self.warning.code != WarningCode.CUBE_LEARN_MODE_STARTED:
                # In learn mode, incomlete nodes are sent.
                await self._update_or_create_node(msg["node"], self.warning.code)
        elif msg_type == "nodes":
            for data in msg["nodes"]:
                await self._update_or_create_node(data, self.warning.code)
        elif msg_type == "relationship":
            self._update_or_create_relationship(msg["relationship"])
        elif msg_type == "relationships":
            self._update_or_create_relationships(msg["relationships"])
            self._remap_relationships()
        elif msg_type == "user":
            self._update_or_create_user(msg["user"])
        elif msg_type == "users":
            for data in msg["users"]:
                self._update_or_create_user(data)
        elif msg_type == "homeegram":
            self._update_or_create_homeegram(msg["homeegram"])
        elif msg_type == "homeegrams":
            for data in msg["homeegrams"]:
                self._update_or_create_homeegram(data)
        elif msg_type == "warning":
            await self._update_warning(msg["warning"])
        else:
            _LOGGER.debug(
                "Unknown/Unsupported message type: %s.\nMessage: %s", msg_type, msg
            )

        await self.on_message(msg)

    async def _handle_attribute_change(self, attribute_data: dict) -> None:
        """Handle an attribute changed message."""

        _LOGGER.debug("Updating attribute %s", attribute_data["id"])

        attr_node_id = attribute_data["node_id"]
        node = self.get_node_by_id(attr_node_id)
        if node is not None:
            node.update_attribute(attribute_data)
            await self.on_attribute_updated(attribute_data, node)

    async def _update_or_create_node(
        self, node_data: dict, warning_code: WarningCode
    ) -> None:
        existing_node = self.get_node_by_id(node_data["id"])
        if (
            existing_node is not None
            and not node_data["attributes"]
            and warning_code == WarningCode.CUBE_REMOVE_MODE_STARTED
        ):
            # A node without attributes is deleted. checking for remove mode for security.
            _LOGGER.debug(
                "Node %s has no attributes and cube remove mode is active. Removing",
                existing_node.id,
            )
            self.nodes.remove(existing_node)
            self._remap_relationships()
            for listener in self._nodes_listeners:
                await listener(existing_node, False)
            return
        if existing_node is not None:
            existing_node.set_data(node_data)
        else:
            self.nodes.append(HomeeNode(node_data))
            self._remap_relationships()
            _LOGGER.debug("Notifying listener of new node %s", self.nodes[-1].id)
            for listener in self._nodes_listeners:
                await listener(self.nodes[-1], True)

    def _update_or_create_group(self, data: dict) -> None:
        group = self.get_group_by_id(data["id"])
        if group is not None:
            group.set_data(data)
        else:
            self.groups.append(HomeeGroup(data))
            self._remap_relationships()

    def _update_or_create_relationship(self, data: dict) -> None:
        relationship: HomeeRelationship | None = next(
            (r for r in self.relationships if r.id == data["id"]), None
        )

        if relationship is not None:
            relationship.set_data(data)
        else:
            self.relationships.append(HomeeRelationship(data))
        self._remap_relationships()

    def _update_or_create_relationships(self, data: dict) -> None:
        if len(self.relationships) <= 0:
            self.relationships = [
                HomeeRelationship(relationship_data) for relationship_data in data
            ]
        else:
            for relationship_data in data:
                self._update_or_create_relationship(relationship_data)

    def _remap_relationships(self) -> None:
        """Remap the relationships between nodes and groups defined by the relationships list."""

        # Clear existing relationships
        for n in self.nodes:
            n.groups.clear()
        for g in self.groups:
            g.nodes.clear()

        for r in self.relationships:
            node = self.get_node_by_id(r.node_id)
            group = self.get_group_by_id(r.group_id)

            if node is not None and group is not None:
                node.groups.append(group)
                group.nodes.append(node)

    def _update_or_create_user(self, data: dict) -> None:
        """Create a user or update if already exists."""
        user = self.get_user_by_id(data["id"])
        if user is not None:
            user.set_data(data)
        else:
            self.users.append(HomeeUser(data))

        # Create / Update the devices of the user
        for device in data["devices"]:
            self._update_or_create_device(device)

    def _update_or_create_device(self, data: dict) -> None:
        """Create a device or update if already exists."""
        device = self.get_device_by_id(data["id"])
        if device is not None:
            device.set_data(data)
        else:
            self.devices.append(HomeeDevice(data))

    def _update_or_create_homeegram(self, data: dict[str, Any]) -> None:
        """Create a homeegram or update if already exists."""
        homeegram = self.get_homeegram_by_id(data["id"])
        if homeegram is not None:
            homeegram.set_data(data)
        else:
            self.homeegrams.append(HomeeGram(data))

    async def _update_warning(self, data: dict) -> None:
        """Set the warning to the latest one received."""
        self.warning.set_data(data)
        await self.on_warning()

    def get_node_index(self, node_id: int) -> int:
        """Return the index of the node with the given id or -1 if none exists."""
        return next((i for i, node in enumerate(self.nodes) if node.id == node_id), -1)

    def get_node_by_id(self, node_id: int) -> HomeeNode | None:
        """Return the node with the given id or 'None' if none exists."""
        index = self.get_node_index(node_id)
        return self.nodes[index] if index != -1 else None

    def get_group_index(self, group_id: int) -> int:
        """Return the index of the group with the given id or -1 if none exists."""
        return next(
            (i for i, group in enumerate(self.groups) if group.id == group_id), -1
        )

    def get_group_by_id(self, group_id: int) -> HomeeGroup | None:
        """Return the group with the given id or 'None' if no group with the given id exists."""
        index = self.get_group_index(group_id)
        return self.groups[index] if index != -1 else None

    def get_user_by_id(self, user_id: int) -> HomeeUser | None:
        """Return the user with the given id or 'None' if no user with the given id exists."""
        index = next((i for i, user in enumerate(self.users) if user.id == user_id), -1)
        return self.users[index] if index != -1 else None

    def get_device_by_id(self, device_id: int) -> HomeeDevice | None:
        """Return the device with the given id or 'None' if no device with the given id exists."""
        index = next(
            (i for i, device in enumerate(self.devices) if device.id == device_id), -1
        )
        return self.devices[index] if index != -1 else None

    def get_homeegram_by_id(self, homeegram_id: int) -> HomeeGram | None:
        """Return the homeegram with the given id or 'None' if none with the given id exists."""
        index = next(
            (
                i
                for i, homeegram in enumerate(self.homeegrams)
                if homeegram.id == homeegram_id
            ),
            -1,
        )
        return self.homeegrams[index] if index != -1 else None

    async def set_value(self, device_id: int, attribute_id: int, value: float) -> None:
        """Set the target value of an attribute of a device."""

        _LOGGER.debug(
            "Set value: Device: %s Attribute: %s To: %s", device_id, attribute_id, value
        )
        await self.send(
            f"PUT:/nodes/{device_id}/attributes/{attribute_id}?target_value={value}"
        )

    async def update_node(self, node_id: int) -> None:
        """Request current data for a node."""
        _LOGGER.debug("Request current data for node %s", node_id)
        await self.send(f"GET:/nodes/{node_id}/")

    async def update_attribute(self, node_id: int, attribute_id: int) -> None:
        """Request current data for an attribute."""
        _LOGGER.debug(
            "Request current data for attribute %s of device %s", attribute_id, node_id
        )
        await self.send(f"GET:/nodes/{node_id}/attributes/{attribute_id}")

    async def play_homeegram(self, homeegram_id: int) -> None:
        """Invoke a homeegram."""

        await self.send(f"PUT:homeegrams/{homeegram_id}?play=1")

    @property
    def url(self) -> str:
        """Local homee url."""

        return f"http://{self.host}:7681"

    @property
    def ws_url(self) -> str:
        """Local homee websocket url."""

        return f"ws://{self.host}:7681"

    async def wait_until_connected(self) -> Literal[True]:
        """Return a coroutine that runs until a connection has been established."""
        return await self._connected_event.wait()

    async def wait_until_disconnected(self) -> Literal[True]:
        """Return a coroutine that runs until the connection has been closed."""
        return await self._disconnected_event.wait()

    async def on_reconnect(self) -> None:
        """Execute right before a reconnection attempt is started."""
        _LOGGER.debug("Homee %s Reconnecting", self.device)

    async def on_max_retries(self) -> None:
        """Execute if the maximum amount of retries was reached."""
        _LOGGER.info(
            "Could not reconnect Homee %s after %s retries",
            self.device,
            self.max_retries,
        )

    async def on_connected(self) -> None:
        """Execute once the websocket connection has been established."""
        for listener in self._connection_listeners:
            await listener(True)
        if self.retries > 0:
            _LOGGER.debug(
                "Homee %s Reconnected after %s retries", self.device, self.retries
            )
            self.retries = 0

    async def on_disconnected(self, error: Exception | None = None) -> None:
        """Execute after the websocket connection has been closed."""
        if not self.should_close:
            for listener in self._connection_listeners:
                await listener(False)

            _LOGGER.info("Homee %s Disconnected. Error: %s", self.device, error)

    async def on_error(self, error: Exception | None = None) -> None:
        """Execute after an error has occurred."""
        _LOGGER.info("An error occurred: %s", error)

    async def on_message(self, msg: dict) -> None:
        """Execute when the websocket receives a message.

        The message is automatically parsed from json into a dictionary.
        """

    async def on_warning(self) -> None:
        """Execute when a warning message is received."""
        if self.warning.code == WarningCode.CUBE_LEARN_MODE_SUCCESSFUL:
            # we need to get all nodes again, since there is no way to get the newest only.
            await self.send("GET:/nodes/")

    async def on_attribute_updated(self, attribute_data: dict, node: HomeeNode) -> None:
        """Execute when an 'attribute' message was received and an attribute was updated.

        Contains the parsed json attribute data and the corresponding node instance.
        """


class HomeeException(Exception):
    """Base class for all errors thrown by this library."""

    def __init__(self, reason: str | None = None) -> None:
        self.reason = reason


class HomeeConnectionFailedException(HomeeException):
    """Raised if connection can not be established."""


class HomeeAuthFailedException(HomeeException):
    """Raised if no valid access token could be acquired."""



"""Data model for Homees various data items."""

from enum import IntEnum

from collections.abc import Callable
from typing import Any, Self, Type
import logging
import re
from urllib.parse import unquote
from .const import (
    AttributeBasedOn,
    AttributeChangedBy,
    AttributeState,
    AttributeType,
    DeviceApp,
    DeviceOS,
    DeviceType,
    GroupCategory,
    NodeProfile,
    NodeProtocol,
    NodeState,
    UserRole,
    WarningCode,
)

_LOGGER = logging.getLogger(__name__)


def log_unknown_value(enum_type: Type[IntEnum], value: int) -> None:
    """Log a warning if a value does not exist in an enum."""
    _LOGGER.warning(
        (
            "Unknown %s %s. Please submit a bug report at "
            "https://github.com/Taraman17/pyHomee/issues"
        ),
        enum_type.__name__,
        value,
    )


class HomeeObject:
    """Base class for Homee objects."""

    def __init__(self, data: dict[str, Any]) -> None:
        """Initialize the object."""
        self._data = data
        self.on_changed_listeners: list[Callable] = []

    @property
    def raw_data(self) -> dict[str, Any]:
        """Return the raw dict of the object."""
        return self._data

    def set_data(self, data: dict[str, Any]) -> None:
        """Update data of the attribute."""
        self._data = data

        for listener in self.on_changed_listeners:
            listener(self)

    def add_on_changed_listener(
        self, listener: Callable[[Self], None]
    ) -> Callable[[], None]:
        """Add on_changed listener to node."""
        self.on_changed_listeners.append(listener)

        def remove_listener() -> None:
            self.on_changed_listeners.remove(listener)

        return remove_listener


class HomeeAttributeOptions:
    """Representation of attributes options."""

    def __init__(self, attribute_options: dict[str, Any]) -> None:
        """Initialize options."""
        self._data = attribute_options

    @property
    def can_observe(self) -> list[int]:
        """List (int) of attribute types that this attribute can observe."""
        if "can_observe" in self._data:
            return list(self._data["can_observe"])

        return []

    @property
    def observes(self) -> list[int]:
        """List (int) of attribute ids that this attribute observes."""
        if "observes" in self._data:
            return list(self._data["observes"])

        return []

    @property
    def observed_by(self) -> list[int]:
        """List (int) of attribute ids that observe this attribute."""
        if "observed_by" in self._data:
            return list(self._data["observed_by"])

        return []

    @property
    def automations(self) -> list[int]:
        """List (str) of automations for thie attribute."""
        if "automations" in self._data:
            return list(self._data["automations"])

        return []

    @property
    def history(self) -> dict[str, int | bool] | None:
        """History data for the attribute.

        {'day': int, 'week': int, 'month': int, 'stepped': bool}.
        """
        if "history" in self._data:
            return dict(self._data["history"])

        return None

    @property
    def reverse_control_ui(self) -> bool:
        """Do up/down controls work in opposite direction."""
        if "reverse_control_ui" in self._data:
            return bool(self._data["reverse_control_ui"])

        return False


class HomeeAttribute(HomeeObject):
    """Representation of a Homee attribute."""

    @property
    def id(self) -> int:
        """The unique id of the attribute."""
        return int(self._data["id"])

    @property
    def node_id(self) -> int:
        """The id of the node the attribute belongs to."""
        return int(self._data["node_id"])

    @property
    def instance(self) -> int:
        """If more than one attribute of same type is present, they are numbered starting at 1."""
        return int(self._data["instance"])

    @property
    def minimum(self) -> float:
        """The minimum possible value of the attribute."""
        return float(self._data["minimum"])

    @property
    def maximum(self) -> float:
        """The maximum possible value of the attribute."""
        return float(self._data["maximum"])

    @property
    def current_value(self) -> float:
        """The current value of the attribute."""
        return float(self._data["current_value"])

    @property
    def target_value(self) -> float:
        """The target value of the attribute.

        Only used to change the attribute value.
        In most cases you want to use current_value instead.
        """
        return float(self._data["target_value"])

    @property
    def last_value(self) -> float:
        """The last value of the attribute. In most cases you want to use current_value instead."""
        return float(self._data["last_value"])

    @property
    def unit(self) -> str:
        """The decoded unit of the attribute."""
        return unquote(self._data["unit"])

    @property
    def step_value(self) -> float:
        """The step value used for attributes with discret increments."""
        return float(self._data["step_value"])

    @property
    def editable(self) -> bool:
        """Wether the attribute is editable of read only."""
        return bool(self._data["editable"])

    @property
    def type(self) -> AttributeType:
        """The attribute type. Compare with const.AttributeType."""
        try:
            return AttributeType(self._data["type"])
        except ValueError:
            log_unknown_value(AttributeType, self._data["type"])

        return AttributeType.NONE

    @property
    def state(self) -> AttributeState:
        """The attribute state. Compare with const.AttributeState."""
        try:
            return AttributeState(self._data["state"])
        except ValueError:
            log_unknown_value(AttributeState, self._data["state"])

        return AttributeState.NONE

    @property
    def last_changed(self) -> int:
        """The last time the attribute was changed."""
        return int(self._data["last_changed"])

    @property
    def changed_by(self) -> AttributeChangedBy:
        """How the attribute was changed. Compare with const.AttributeChangedBy."""
        try:
            return AttributeChangedBy(self._data["changed_by"])
        except ValueError:
            log_unknown_value(AttributeChangedBy, self._data["changed_by"])

        return AttributeChangedBy.NONE

    @property
    def changed_by_id(self) -> int:
        """The id of the user/homeegram the attribute was changed by."""
        return int(self._data["changed_by_id"])

    @property
    def based_on(self) -> AttributeBasedOn:
        """TODO"""
        try:
            return AttributeBasedOn(self._data["based_on"])
        except ValueError:
            log_unknown_value(AttributeBasedOn, self._data["based_on"])

        return AttributeBasedOn.NONE

    @property
    def name(self) -> str:
        """The decoded name of the attribute."""
        return unquote(self._data["name"])

    @property
    def data(self) -> str:
        """The data string of the attribute. Note that the data may be uri encoded."""
        return unquote(self._data["data"])

    @property
    def options(self) -> HomeeAttributeOptions | None:
        """The options collection of the attribute. Optional, not on every attribute."""
        if "options" in self._data:
            return HomeeAttributeOptions(self._data["options"])

        return None

    @property
    def is_reversed(self) -> bool:
        """Check if movement direction is reversed."""
        if self.options is not None:
            return bool(self.options.reverse_control_ui)

        return False

    def get_value(self) -> float | str:
        """Get the current value or data of the attribute."""
        # If the unit of the attribute is 'text', it is stored in .data
        if self.unit == "text":
            return self.data

        return self.current_value


class HomeeNode(HomeeObject):
    """Representation of a node in Homee."""

    def __init__(self, data: dict[str, Any]) -> None:
        """Initialize a Homee node."""
        super().__init__(data)
        self.attributes: list[HomeeAttribute] = []
        for a in self.attributes_raw:
            new_attribute = HomeeAttribute(a)
            new_attribute.add_on_changed_listener(self._on_attribute_changed)
            self.attributes.append(new_attribute)
        self._attribute_map: dict[AttributeType, HomeeAttribute] = {}
        self.remap_attributes()
        self.groups: list[HomeeGroup] = []

    @property
    def id(self) -> int:
        """The unique id of the node."""
        return int(self._data["id"])

    @property
    def name(self) -> str:
        """The decoded primary name of the node."""
        return unquote(self._data["name"])

    @property
    def profile(self) -> NodeProfile:
        """The NodeProfile of this node."""
        try:
            return NodeProfile(self._data["profile"])
        except ValueError:
            log_unknown_value(NodeProfile, self._data["profile"])

        return NodeProfile.NONE

    @property
    def image(self) -> str:
        return unquote(self._data["image"])

    @property
    def favorite(self) -> int:
        return int(self._data["favorite"])

    @property
    def order(self) -> int:
        return int(self._data["order"])

    @property
    def protocol(self) -> NodeProtocol:
        """The network protocol of the node."""
        try:
            return NodeProtocol(self._data["protocol"])
        except ValueError:
            log_unknown_value(NodeProfile, self._data["protocol"])

        return NodeProtocol.NONE

    @property
    def routing(self) -> bool:
        return bool(self._data["routing"])

    @property
    def state(self) -> NodeState:
        """State of availability."""
        try:
            return NodeState(self._data["state"])
        except ValueError:
            log_unknown_value(NodeState, self._data["state"])

        return NodeState.NONE

    @property
    def state_changed(self) -> int:
        return int(self._data["state_changed"])

    @property
    def added(self) -> int:
        return int(self._data["added"])

    @property
    def history(self) -> int:
        return int(self._data["history"])

    @property
    def cube_type(self) -> int:
        """Type of the Homee cube the node is part of."""
        return int(self._data["cube_type"])

    @property
    def note(self) -> str:
        """Text Note describing the node."""
        return unquote(self._data["note"])

    @property
    def services(self) -> int:
        return int(self._data["services"])

    @property
    def phonetic_name(self) -> str:
        """Name of the node."""
        return unquote(self._data["phonetic_name"])

    @property
    def owner(self) -> int:
        return int(self._data["owner"])

    @property
    def security(self) -> int:
        return int(self._data["security"])

    @property
    def attribute_map(self) -> dict[AttributeType, HomeeAttribute] | None:
        """Dict containing all attributes with attributeType as key."""
        return self._attribute_map

    @property
    def attributes_raw(self) -> list[dict[str, Any]]:
        """Return raw dict all the node's attributes."""
        return list(self._data["attributes"])

    def set_data(self, data: dict[str, Any]) -> None:
        """Update data of the node."""
        super().set_data(data)
        self.update_attributes(self._data["attributes"])

    def get_attribute_index(self, attribute_id: int) -> int:
        """Find and return attribute for a given index.

        Returns -1 if not found."""
        return next(
            (i for i, a in enumerate(self.attributes) if a.id == attribute_id), -1
        )

    def get_attribute_by_type(
        self, attribute_type: int, instance: int = 0
    ) -> HomeeAttribute | None:
        """Find and return attribute by attributeType.

        If multiple attributes of the same type are present,
        the instance number can be used to select the correct one."""
        for a in self.attributes:
            if a.type == attribute_type and a.instance == instance:
                return a

        return None

    def get_attribute_by_id(self, attribute_id: int) -> HomeeAttribute | None:
        """Find and return attribute for a given id."""
        index = self.get_attribute_index(attribute_id)
        return self.attributes[index] if index != -1 else None

    def _on_attribute_changed(self, attribute: HomeeAttribute) -> None:
        for listener in self.on_changed_listeners:
            listener(self)

    def update_attribute(self, attribute_data: dict[str, Any]) -> None:
        """Update a single attribute of a node."""
        attribute = self.get_attribute_by_id(attribute_data["id"])
        if attribute is not None:
            attribute.set_data(attribute_data)

    def update_attributes(self, attributes: list[dict]) -> None:
        """Update the given attributes."""
        for attr in attributes:
            self.update_attribute(attr)

    def remap_attributes(self) -> None:
        """Remap the node's attributes."""
        self._attribute_map.clear()
        for a in self.attributes:
            self._attribute_map.update({a.type: a})


class HomeeGroup(HomeeObject):
    """Representation of a Homee group."""

    def __init__(self, data: dict[str, Any]) -> None:
        """Initialize a Homee group."""
        super().__init__(data)
        self.nodes: list[HomeeNode] = []

    @property
    def id(self) -> int:
        """Id of the group, unique in Homee."""
        return int(self._data["id"])

    @property
    def name(self) -> str:
        """Decoded user given name of the group."""
        return unquote(self._data["name"])

    @property
    def image(self) -> str:
        return unquote(self._data["image"])

    @property
    def order(self) -> int:
        return int(self._data["order"])

    @property
    def added(self) -> int:
        return int(self._data["added"])

    @property
    def state(self) -> int:
        return int(self._data["state"])

    @property
    def category(self) -> GroupCategory:
        try:
            return GroupCategory(self._data["category"])
        except ValueError:
            log_unknown_value(GroupCategory, self._data["category"])

        return GroupCategory.NONE

    @property
    def phonetic_name(self) -> str:
        return unquote(self._data["phonetic_name"])

    @property
    def note(self) -> str:
        """Note describing the group."""
        return unquote(self._data["note"])

    @property
    def services(self) -> int:
        return int(self._data["services"])

    @property
    def owner(self) -> int:
        return int(self._data["owner"])


class HomeeSettings(HomeeObject):
    """Representation of the settings object passed by Homee."""

    @property
    def address(self) -> str:
        """Street set by user."""
        return unquote(self._data["address"])

    @property
    def city(self) -> str:
        """City set by user."""
        return unquote(self._data["city"])

    @property
    def zip(self) -> str:
        """Zip code set by user."""
        return unquote(self._data["zip"])

    @property
    def state(self) -> str:
        """State set by user."""
        return unquote(self._data["state"])

    @property
    def latitude(self) -> float:
        """Latitude of set position of Homee."""
        return float(self._data["latitude"])

    @property
    def longitude(self) -> float:
        """Longitude of set position of Homee."""
        return float(self._data["longitude"])

    @property
    def country(self) -> str:
        """Country set by user."""
        return unquote(self._data["country"])

    @property
    def language(self) -> str:
        """Frontend language."""
        return unquote(self._data["language"])

    @property
    def remote_access(self) -> int:
        """Remote access enabled or not."""
        return int(self._data["remote_access"])

    @property
    def beta(self) -> bool:
        """Is user accepting beta releases of firmware."""
        return bool(self._data["beta"])

    @property
    def webhooks_key(self) -> str:
        """Key used for webhooks."""
        return str(self._data["webhooks_key"])

    @property
    def automatic_location_detection(self) -> bool:
        return bool(self._data["automatic_location_detection"])

    @property
    def polling_interval(self) -> float:
        """Standard polling interval set in Homee."""
        return float(self._data["polling_interval"])

    @property
    def timezone(self) -> str:
        """Timezone of Homee."""
        return unquote(self._data["timezone"])

    @property
    def enable_analytics(self) -> bool:
        """Send analytical data back home."""
        return bool(self._data["enable_analytics"])

    @property
    def homee_name(self) -> str:
        """Decoded name of Homee."""
        return unquote(self._data["homee_name"])

    @property
    def LastMissingCubeNotification(self) -> str:
        return unquote(self._data["LastMissingCubeNotification"])

    @property
    def local_ssl_enabled(self) -> bool:
        return bool(self._data["local_ssl_enabled"])

    @property
    def wlan_enabled(self) -> bool:
        return bool(self._data["wlan_enabled"])

    @property
    def wlan_ssid(self) -> str:
        return str(self._data["wlan_ssid"])

    @property
    def wlan_mode(self) -> int:
        return int(self._data["wlan_mode"])

    @property
    def mac_address(self) -> str:
        """Return MAC Address derived from HomeeID"""
        return ":".join(re.findall("..", self._data["uid"]))

    @property
    def internet_access(self) -> bool:
        return bool(self._data["internet_access"])

    @property
    def lan_enabled(self) -> bool:
        return bool(self._data["lan_enabled"])

    @property
    def lan_ip_address(self) -> str:
        return unquote(self._data["lan_ip_address"])

    @property
    def available_ssids(self) -> list[str]:
        return list(self._data["available_ssids"])

    @property
    def time(self) -> int:
        return int(self._data["time"])

    @property
    def civil_time(self) -> str:
        return str(self._data["civil_time"])

    @property
    def version(self) -> str:
        return str(self._data["version"])

    @property
    def uid(self) -> str:
        return str(self._data["uid"])

    @property
    def cubes(self) -> list[dict[str, Any]]:
        """List of cubes attached to this Homee."""
        return list(self._data["cubes"])

    @property
    def extensions(self) -> list[dict]:
        return list(self._data["extensions"])


class HomeeRelationship(HomeeObject):
    """Representation of a Homee relationship."""

    @property
    def id(self) -> int:
        """Id unique to this Homee."""
        return int(self._data["id"])

    @property
    def group_id(self) -> int:
        return int(self._data["group_id"])

    @property
    def node_id(self) -> int:
        return int(self._data["node_id"])

    @property
    def homeegram_id(self) -> int:
        return int(self._data["homeegram_id"])

    @property
    def order(self) -> int:
        return int(self._data["order"])

    def set_data(self, data: dict[str, Any]) -> None:
        """Update data of the relationship."""
        self._data = data
        for listener in self.on_changed_listeners:
            listener(self)

    def add_on_changed_listener(
        self, listener: Callable[[Self], None]
    ) -> Callable[[], None]:
        """Add on_changed listener to node."""
        self.on_changed_listeners.append(listener)

        def remove_listener() -> None:
            self.on_changed_listeners.remove(listener)

        return remove_listener


class HomeeWarningData:
    """Representation of the data part of a Homee warning."""

    def __init__(self, data: dict[str, Any]) -> None:
        """Initialize warning data."""
        self._data = data

    @property
    def protocol(self) -> NodeProtocol | None:
        """Return the protocol, the warning originates from."""
        if "protocol" in self._data:
            try:
                return NodeProtocol(self._data["protocol"])
            except ValueError:
                log_unknown_value(NodeProtocol, self._data["protocol"])

            return NodeProtocol.NONE

        return None

    @property
    def protocol_string(self) -> str:
        """Return the descriptive string for the protocol."""
        if "protocol" in self._data:
            try:
                return NodeProtocol(self._data["protocol"]).name
            except ValueError:
                log_unknown_value(NodeProtocol, self._data["protocol"])

            return NodeProtocol.NONE.name

        return ""

    @property
    def reason(self) -> str:
        """Return the reason for the warning."""
        if "reason" in self._data:
            return unquote(self._data["reason"])

        return ""


class HomeeWarning(HomeeObject):
    """Representation of a Homee warning message."""

    @property
    def code(self) -> WarningCode:
        """Return the numerical code of the warning."""
        try:
            return WarningCode(self._data["code"])
        except ValueError:
            log_unknown_value(WarningCode, self._data["code"])

        return WarningCode.NONE

    @property
    def code_string(self) -> str:
        """Return the descriptive string for the warning code."""
        try:
            return WarningCode(self._data["code"]).name
        except ValueError:
            log_unknown_value(WarningCode, self._data["code"])

        return WarningCode.NONE.name

    @property
    def description(self) -> str:
        """Return the text description of the warning."""
        return unquote(self._data["description"])

    @property
    def message(self) -> str:
        """Return the message of the warning."""
        return unquote(self._data["message"])

    @property
    def data(self) -> HomeeWarningData | None:
        """The data collection of the warning. Optional, not on every warning."""
        if "data" in self._data:
            return HomeeWarningData(self._data["data"])

        return None


class HomeeDevice(HomeeObject):
    """Represent a Homee device."""

    @property
    def id(self) -> int:
        """Return the unique id of the device."""
        return int(self._data["id"])

    @property
    def user_id(self) -> int:
        """Return the id of the user the device belongs to."""
        return int(self._data["user_id"])

    @property
    def hardware_id(self) -> str:
        """Return the hardware id of the device."""
        return str(self._data["hardware_id"])

    @property
    def name(self) -> str:
        """Return the name of the device."""
        return unquote(self._data["name"])

    @property
    def added(self) -> int:
        """Return the time the device was added."""
        return int(self._data["added"])

    @property
    def last_connected(self) -> int:
        """Return the last time the device was connected."""
        return int(self._data["last_connected"])

    @property
    def os(self) -> DeviceOS:
        """Return the operating system of the device."""
        try:
            return DeviceOS(self._data["os"])
        except ValueError:
            log_unknown_value(DeviceOS, self._data["os"])

        return DeviceOS.NONE

    @property
    def type(self) -> DeviceType:
        """Return the type of the device."""
        try:
            return DeviceType(self._data["type"])
        except ValueError:
            log_unknown_value(DeviceType, self._data["type"])

        return DeviceType.NONE

    @property
    def app(self) -> DeviceApp:
        """Return the app version of the device."""
        try:
            return DeviceApp(self._data["app"])
        except ValueError:
            log_unknown_value(DeviceApp, self._data["app"])

        return DeviceApp.NONE

    @property
    def connected(self) -> bool:
        """Return whether the device is currently connected."""
        return bool(self._data["connected"])

    @property
    def push_registration_id(self) -> str:
        """Return the push registration id of the device."""
        return str(self._data["push_registration_id"])


class HomeeUser(HomeeObject):
    """Represent a Homee user."""

    @property
    def id(self) -> int:
        """Return the unique id of the user."""
        return int(self._data["id"])

    @property
    def username(self) -> str:
        """Return the username of the user."""
        return str(self._data["username"])

    @property
    def forename(self) -> str:
        """Return the forename of the user."""
        return unquote(self._data["forename"])

    @property
    def surname(self) -> str:
        """Return the surname of the user."""
        return unquote(self._data["surname"])

    @property
    def image(self) -> str:
        """Return the image of the user."""
        return unquote(self._data["image"])

    @property
    def role(self) -> UserRole:
        """Return the role of the user."""
        try:
            return UserRole(self._data["role"])
        except ValueError:
            log_unknown_value(UserRole, self._data["role"])

        return UserRole.NONE

    @property
    def type(self) -> int:
        """Return the type of the user."""
        return int(self._data["type"])

    @property
    def email(self) -> str:
        """Return the email of the user."""
        return unquote(self._data["email"])

    @property
    def phone(self) -> str:
        """Return the phone number of the user."""
        return str(self._data["phone"])

    @property
    def added(self) -> int:
        """Return the time the user was added."""
        return int(self._data["added"])

    @property
    def homee_image(self) -> str:
        """Return the homee image of the user."""
        return unquote(self._data["homee_image"])

    @property
    def access(self) -> int:
        """Return the access level of the user."""
        return int(self._data["access"])

    @property
    def presence_detection(self) -> bool:
        """Return whether presence detection is enabled for the user."""
        return bool(self._data["presence_detection"])

    @property
    def cube_push_notifications(self) -> bool:
        """Return whether cube push notifications are enabled for the user."""
        return bool(self._data["cube_push_notifications"])

    @property
    def cube_email_notifications(self) -> bool:
        """Return whether cube email notifications are enabled for the user."""
        return bool(self._data["cube_email_notifications"])

    @property
    def cube_sms_notifications(self) -> bool:
        """Return whether cube SMS notifications are enabled for the user."""
        return bool(self._data["cube_sms_notifications"])

    @property
    def warning_push_notifications(self) -> bool:
        """Return whether warning push notifications are enabled for the user."""
        return bool(self._data["warning_push_notifications"])

    @property
    def warning_push_notifications_as_critical(self) -> bool:
        """Return whether warning push notifications are marked as critical for the user."""
        return bool(self._data["warning_push_notifications_as_critical"])

    @property
    def warning_email_notifications(self) -> bool:
        """Return whether warning email notifications are enabled for the user."""
        return bool(self._data["warning_email_notifications"])

    @property
    def warning_sms_notifications(self) -> bool:
        """Return whether warning SMS notifications are enabled for the user."""
        return bool(self._data["warning_sms_notifications"])

    @property
    def node_push_notifications(self) -> bool:
        """Return whether node push notifications are enabled for the user."""
        return bool(self._data["node_push_notifications"])

    @property
    def node_email_notifications(self) -> bool:
        """Return whether node email notifications are enabled for the user."""
        return bool(self._data["node_email_notifications"])

    @property
    def node_sms_notifications(self) -> bool:
        """Return whether node SMS notifications are enabled for the user."""
        return bool(self._data["node_sms_notifications"])

    @property
    def update_push_notifications(self) -> bool:
        """Return whether update push notifications are enabled for the user."""
        return bool(self._data["update_push_notifications"])

    @property
    def update_email_notifications(self) -> bool:
        """Return whether update email notifications are enabled for the user."""
        return bool(self._data["update_email_notifications"])

    @property
    def update_sms_notifications(self) -> bool:
        """Return whether update SMS notifications are enabled for the user."""
        return bool(self._data["update_sms_notifications"])

    @property
    def homeegram_push_notifications(self) -> bool:
        """Return whether homeegram push notifications are enabled for the user."""
        return bool(self._data["homeegram_push_notifications"])

    @property
    def homeegram_email_notifications(self) -> bool:
        """Return whether homeegram email notifications are enabled for the user."""
        return bool(self._data["homeegram_email_notifications"])

    @property
    def homeegram_sms_notifications(self) -> bool:
        """Return whether homeegram SMS notifications are enabled for the user."""
        return bool(self._data["homeegram_sms_notifications"])

    @property
    def api_push_notifications(self) -> bool:
        """Return whether API push notifications are enabled for the user."""
        return bool(self._data["api_push_notifications"])

    @property
    def api_email_notifications(self) -> bool:
        """Return whether API email notifications are enabled for the user."""
        return bool(self._data["api_email_notifications"])

    @property
    def api_sms_notifications(self) -> bool:
        """Return whether API SMS notifications are enabled for the user."""
        return bool(self._data["api_sms_notifications"])

    @property
    def plan_push_notifications(self) -> bool:
        """Return whether plan push notifications are enabled for the user."""
        return bool(self._data["plan_push_notifications"])

    @property
    def plan_email_notifications(self) -> bool:
        """Return whether plan email notifications are enabled for the user."""
        return bool(self._data["plan_email_notifications"])

    @property
    def plan_sms_notifications(self) -> bool:
        """Return whether plan SMS notifications are enabled for the user."""
        return bool(self._data["plan_sms_notifications"])

    @property
    def watchdog_push_notifications(self) -> bool:
        """Return whether watchdog push notifications are enabled for the user."""
        return bool(self._data["watchdog_push_notifications"])

    @property
    def watchdog_email_notifications(self) -> bool:
        """Return whether watchdog email notifications are enabled for the user."""
        return bool(self._data["watchdog_email_notifications"])

    @property
    def watchdog_sms_notifications(self) -> bool:
        """Return whether watchdog SMS notifications are enabled for the user."""
        return bool(self._data["watchdog_sms_notifications"])

    @property
    def devices(self) -> list[HomeeDevice]:
        """Return the list of devices associated with the user."""
        devices = []
        for device in self._data["devices"]:
            devices.append(HomeeDevice(device))

        return devices
